#ifndef bqtExprHH
#define bqtExprHH

//#define BINDING_DEBUG

#include <string>
#include <vector>
#include <functional>

#include <memory>

typedef std::unique_ptr<class expr> exprp;

#include "tristate"
#include "tokens.hh"
#include "symtables.hh"
#include "lexer.hh"
#include "types.hh"

enum ExprCompileDefs
{
    exprdef_nothingparticular,
    exprdef_may_produce_char_from_string,
    exprdef_cast_to_int,
    exprdef_membername,
    exprdef_pure_condition
};


/* std::function cannot include default parameters */
class expr_visitor: public std::function<void(
                        const expr*, tristate write_target/* = false*/,
                        const std::string& name/*=""*/, unsigned param_index/*=0*/)>
{
    typedef std::function<void(const expr*,tristate,const std::string&,unsigned)> parent;
public:
    template<typename T>
    expr_visitor(const T& b) : parent(b) { }

    inline void operator() (const expr* e,tristate w=false,const std::string& n="",unsigned p=0) const
        { parent::operator() (e, w, n,p); }
};


class expr
{
public:
    expr();
    virtual ~expr();

    tristate is_num() const;
    tristate is_string() const;
    virtual TypeP GetType() const = 0;
    virtual bool is_const() const { return false; }

    virtual bool MayErr() const = 0;
    virtual bool IsComparison() const { return false; }

    virtual expr* optimized() &&
    {
        // FIXME: uh
        return this;
    }

    virtual std::string Compile(ExprCompileDefs defs=exprdef_nothingparticular) const;

    virtual void ForEachSubExpr(const expr_visitor& ) const {  }

    virtual bool IsSameAs(const expr& ) const { return false; }
};

class expr_string: public expr
{
    std::string value;
public:
    expr_string(): value() { }
    expr_string(const char* v) : value(v) { }
    expr_string(const std::string& v) : value(v) { }

    virtual TypeP GetType() const { return TypeStr; }
    virtual bool is_const() const  { return true; }

    const std::string& get() const { return value; }

    virtual std::string Compile(ExprCompileDefs defs) const;
    virtual bool MayErr() const { return false; }

    virtual void ForEachSubExpr(const expr_visitor& func) const;

    virtual bool IsSameAs(const expr& ) const;
};

class expr_number: public expr
{
    typedef long long int itype;
    typedef long double ftype;
    TypeP type;
    itype int_val;
    ftype float_val;
public:
    expr_number() : type(TypeInt), int_val(0), float_val(0) { }
    expr_number(long long int i) : type(), int_val(), float_val() { InitIntImpl(i); }
    expr_number(long int i)      : type(), int_val(), float_val() { InitIntImpl(i); }
    expr_number(unsigned long i) : type(), int_val(), float_val() { InitIntImpl(i); }
    expr_number(int i)           : type(), int_val(), float_val() { InitIntImpl(i); }
    expr_number(long long int i, TypeP t) : type(t),int_val(i),float_val(i) {}
    expr_number(long int i, TypeP t)      : type(t),int_val(i),float_val(i) {}
    expr_number(int i, TypeP t)           : type(t),int_val(i),float_val(i) {}
    expr_number(long double f, TypeP t=TypeDbl)   : type(t),int_val((itype)f),float_val(f) {}
    expr_number(double f, TypeP t=TypeDbl)        : type(t),int_val((itype)f),float_val(f) {}

    virtual TypeP GetType() const { return type; }
    virtual bool is_const() const  { return true; }

    bool is_int() const;
    long double   get_f() const;
    long long int get_i() const;

private:
    void InitIntImpl(long long int i)
    {
        int_val = i;
        if(i >= 0x80000000u)   type = TypeUlong;
        else if(i >= 0x10000u) type = TypeLng;
        else if(i >= 0x8000u)  type = TypeUint;
        else if(i >= 0x100u)   type = TypeInt;
        else if(i >= 0u)       type = TypeUbyte;
        else if(i >= -0x80)    type = TypeByte;
        else if(i >= -0x8000)  type = TypeInt;
        else type = TypeLng;
    }

public:
    virtual std::string Compile(ExprCompileDefs defs) const;
    virtual bool MayErr() const { return false; }

    virtual void ForEachSubExpr(const expr_visitor& func) const;

    virtual bool IsSameAs(const expr& ) const;
};


class expr_unary: public expr
{
public:
    QBtoken token;
    exprp p;
public:
    expr_unary(QBtoken tok, exprp&& a) : token(tok), p(std::move(a)) { }

    virtual TypeP GetType() const;
    QBtoken GetToken() const { return token; }

    virtual expr* optimized() &&;

    virtual std::string Compile(ExprCompileDefs defs) const;

    /* Possible expr_unaries are: NOT and MINUS. Neither of them can produce errors if p cannot. */
    virtual bool MayErr() const { return p->MayErr(); }

    virtual bool IsComparison() const { return token==NOT && p->IsComparison(); }

    virtual void ForEachSubExpr(const expr_visitor& func) const;

    virtual bool IsSameAs(const expr& ) const;
};

class expr_binary: public expr
{
public:
    QBtoken token;
    exprp p0, p1;
public:
    expr_binary(QBtoken tok, exprp&& a, exprp&& b) : token(tok), p0(std::move(a)), p1(std::move(b)) { }

    virtual TypeP GetType() const;
    /* Note: IMP and EQV have been eliminated in the parse phase. */

    virtual expr* optimized() &&;

    virtual std::string Compile(ExprCompileDefs defs) const;

    virtual bool IsComparison() const;

    virtual void ForEachSubExpr(const expr_visitor& func) const;

    exprp& get_p0() { return p0; }
    exprp& get_p1() { return p1; }
    QBtoken GetToken() const { return token; }

    virtual bool MayErr() const;

    virtual bool IsSameAs(const expr& ) const;
};

class expr_label: public expr
{
    std::string label;
public:
    expr_label() : label(0) { }
    expr_label(const std::string& l) : label(l) { }
    expr_label(long long int i) : label( std::to_string(i) ) { }

    virtual TypeP GetType() const { return 0; }

    const std::string& get_label() const { return label; }

    virtual std::string Compile(ExprCompileDefs defs) const;
    virtual bool MayErr() const { return false; }

    virtual void ForEachSubExpr(const expr_visitor& func) const;

    virtual bool IsSameAs(const expr& ) const;
};

class expr_userfun: public expr
{
    LexLocation lexloc;
    std::string name;
    TypeP type;
    std::vector<exprp> params;
public:
    expr_userfun(const LexLocation& l, const std::string& n) : lexloc(l), name(n), type(), params() { }
    expr_userfun(const LexLocation& l,
                 const std::string& n,
                 TypeP t,
                 std::vector<exprp>&& p) : lexloc(l), name(n), type(t), params(std::move(p)) { }

    const std::string& getname() const { return name; }

    virtual TypeP GetType() const { return type; }

    friend class expr_var;

    virtual std::string Compile(ExprCompileDefs defs) const;

    std::string Compile(ExprCompileDefs defs, bool is_statement) const;

    virtual void ForEachSubExpr(const expr_visitor& func) const;

    exprp& get_param(std::size_t a) { return params[a]; }

    virtual bool IsSameAs(const expr& ) const;

    // Calling the function does not produce errors. The function itself
    // may produce errors, but they are not the responsibility of the caller.
    virtual bool MayErr() const
    {
        for(const auto& p: params) if(p->MayErr()) return true;
        return false;
    }
};

class expr_builtin_function: public expr
{
protected:
    LexLocation lexloc;
    QBtoken token;
    std::vector<exprp> params;
    std::vector<bool> lvaluemap;
    std::string pattern;
public:
    expr_builtin_function(const LexLocation& l, QBtoken tok, const std::string& p="")
        : lexloc(l), token(tok), params(), lvaluemap(), pattern(p) { }
    expr_builtin_function(const LexLocation& l, QBtoken tok, exprp&& a, const std::string& p="")
        : lexloc(l), token(tok), params(), lvaluemap(), pattern(p)
        { params.emplace_back(std::move(a)); }

    void set_params(std::vector<exprp>&& p) { params=std::move(p); }
    void set_lvalue_map(const std::vector<bool>& l) { lvaluemap=l; }
    void set_pattern(const std::string& p) { pattern=p; }
    exprp&  GetParam(std::size_t n)            { return params[n]; }
    const exprp& GetParam(std::size_t n) const { return params[n]; }
    std::size_t GetNumParams() const { return params.size(); }

    QBtoken GetToken() const { return token; }
    void ChangeToken(QBtoken newtoken) { token = newtoken; }

    virtual std::string Compile(ExprCompileDefs defs) const;

    virtual void ForEachSubExpr(const expr_visitor& func) const;

    virtual bool IsSameAs(const expr& ) const;

    virtual bool MayErr() const;
};

class expr_builtin_numeric: public expr_builtin_function
{
public:
    template<typename...T>
    expr_builtin_numeric(T&&... args)
        : expr_builtin_function(std::forward<T>(args)...) { }

    virtual TypeP GetType() const;

    virtual expr* optimized() &&;
};

class expr_builtin_string: public expr_builtin_function
{
public:
    template<typename... T>
    expr_builtin_string(T&&... args)
        : expr_builtin_function(std::forward<T>(args)...) { }

    virtual TypeP GetType() const { return TypeStr; }

    virtual expr* optimized() &&;
};

/* These can only occur in select-case */
class expr_caseelse: public expr
{
public:
    expr_caseelse() { }

    virtual TypeP GetType() const { return 0; }

    virtual bool is_const() const { return true; }

    virtual std::string Compile(ExprCompileDefs defs) const;

    virtual bool MayErr() const { return false; }

    virtual void ForEachSubExpr(const expr_visitor& func) const;
};
class expr_caseis: public expr
{
public:
    QBtoken token; // relational comparison operator
    exprp ref;
public:
    expr_caseis(QBtoken t, exprp&& r) : token(t), ref(std::move(r)){ }

    virtual TypeP GetType() const { return 0; }

    virtual bool is_const() const { return ref->is_const(); }

    virtual std::string Compile(ExprCompileDefs defs) const;

    virtual void ForEachSubExpr(const expr_visitor& func) const;

    virtual bool MayErr() const { return ref->MayErr(); }
};
class expr_caserange: public expr
{
public:
    exprp min, max;
public:
    expr_caserange(exprp&& mi, exprp&& ma) : min(std::move(mi)), max(std::move(ma)) { }

    virtual TypeP GetType() const { return 0; }

    virtual bool is_const() const { return min->is_const() && max->is_const(); }

    virtual std::string Compile(ExprCompileDefs defs) const;

    virtual void ForEachSubExpr(const expr_visitor& func) const;

    virtual bool MayErr() const { return min->MayErr() || max->MayErr(); }
};
class expr_caseor: public expr
{
public:
    exprp a,b;
public:
    expr_caseor(exprp&& aa, exprp&& bb) : a(std::move(aa)), b(std::move(bb)) { }

    virtual TypeP GetType() const { return 0; }

    virtual bool is_const() const { return a->is_const() && b->is_const(); }

    virtual std::string Compile(ExprCompileDefs defs) const;

    virtual void ForEachSubExpr(const expr_visitor& func) const;

    virtual bool MayErr() const { return a->MayErr() || b->MayErr(); }
};


class expr_var: public expr
{
    LexLocation lexloc;
    std::string name;
    std::vector<exprp> indexes;
    bool array_as_whole = false;
    std::unique_ptr<expr_var> member;            // NULL if none
    TypeP type;     // Type of referred element (variable, array element, usertype element)
    TypeP basetype; // Type of root element (variable, array itself, usertype itself)
public:
    expr_var(const LexLocation& l, const std::string& n) : lexloc(l), name(n) {}

    virtual TypeP GetType() const { return type; }
    virtual bool is_const() const { return false; }

    void addindex(exprp&& p)    { indexes.emplace_back( std::move(p) ); }
    void addmember(expr_var* m) { member.reset(m); }

    bool hasmember() const { return member.get(); }
    expr_var& getmember() const { return *member.get(); }

    bool iswholearray() const { return array_as_whole; }
    void setwholearray(bool b=true) { array_as_whole = b; }

    bool isarray() const { return iswholearray() || hasindexes(); }

    const std::string& getname() const { return name; }

    bool hasindexes() const { return !indexes.empty(); }
    std::size_t numindexes() const { return indexes.size(); }
    const exprp& getindex(unsigned n) const { return indexes[n]; }

    void set_type(TypeP t)     { type=t; }
    void set_basetype(TypeP t) { basetype=t; }

    expr_userfun* convert_to_functioncall(TypeP type) &&;

    virtual std::string Compile(ExprCompileDefs defs) const;

    /* expr_var may generate errors if:
     *   - Type is a Dynamic array or Param Array  (index() may B_ERROR)
     *   - Type is static array and CheckStaticSubscripts is enabled
     *   - Any of the array indexes may error
     *   - Member can error
     */
    virtual bool MayErr() const;

    virtual void ForEachSubExpr(const expr_visitor& func) const;

    virtual bool IsSameAs(const expr& ) const;
};

class expr_returnvalueholder: public expr_var
{
public :
    explicit expr_returnvalueholder(const LexLocation& lexloc, const std::string& n)
        : expr_var(lexloc, n) { }

    virtual std::string Compile(ExprCompileDefs defs) const;
    virtual bool MayErr() const { return false; }

    virtual void ForEachSubExpr(const expr_visitor& func) const;
};

class expr_lubound: public expr
{
    LexLocation lexloc;
    QBtoken token;
    exprp var;
    exprp num_index;
public:
    expr_lubound(const LexLocation& l, QBtoken t, exprp&& a, exprp&& b = std::unique_ptr<expr>(new expr_number(1)))
        : lexloc(l), token(t), var(std::move(a)), num_index(std::move(b)) { }

    virtual TypeP GetType() const { return TypeLng; }
    virtual bool is_const() const
    {
        if(!num_index->is_const()) return false;
        // Return true if "name" is a $STATIC array, false otherwise.
        if(auto v = dynamic_cast<const expr_var*> (&*var))
            if(v->GetType() && v->GetType()->arraytype == TypeInfo::StaticArray)
                return true; // is const
        return false;
    }

    virtual bool MayErr() const;

    virtual std::string Compile(ExprCompileDefs defs) const;

    virtual void ForEachSubExpr(const expr_visitor& func) const;

    virtual bool IsSameAs(const expr& ) const;
};


static inline void OptimizeExprp(exprp& ptr)
{
    if(!ptr.get()) return;
    expr* e = std::move(*(expr*)ptr.get()).optimized();
    if(e == ptr.get()) return;
    ptr.reset(e);
}

#endif
