#include <vector>
#include <cstdio>
#include <list>
#include <algorithm>
#include <unordered_map>

#include "lexer.hh"
#include "dprintf.hh"
#include "binding.hh"
#include "util.hh"

static const struct
{
    const char* string;
    QBtoken     token;
} ReservedWords_Const[] =
{
    { "ABS"        ,  ABS },
    { "ACCESS"     ,  ACCESS },
    { "ALIAS"      ,  ALIAS },
    { "AND"        ,  AND },
    { "ANY"        ,  ANY },
    { "APPEND"     ,  APPEND },
    { "AS"         ,  AS },
    { "ASC"        ,  ASC },
    { "ATN"        ,  ATN },
    { "BASE"       ,  BASE },
    { "BEEP"       ,  BEEP },
    { "BINARY"     ,  BINARY },
    { "BYVAL"      ,  BYVAL },
    { "CALL"       ,  CALL },
    { "CASE"       ,  CASE },
    { "CDBL"       ,  CDBL },
    { "CDECL"      ,  CDECL },
    { "CHDIR"      ,  CHDIR },
    { "CHR$"       ,  CHR_S },
    { "CINT"       ,  CINT },
    { "CIRCLE"     ,  CIRCLE },
    { "CLEAR"      ,  CLEAR },
    { "CLNG"       ,  CLNG },
    { "CLOSE"      ,  CLOSE },
    { "CLS"        ,  CLS },
    { "COLOR"      ,  COLOR },
    { "COMMAND$"   ,  COMMAND_S },
    { "COMMON"     ,  COMMON },
    { "CONST"      ,  CONST },
    { "COS"        ,  COS },
    { "CSNG"       ,  CSNG },
    { "CSRLIN"     ,  CSRLIN },
    { "CVD"        ,  CVD },
    { "CVI"        ,  CVI },
    { "CVL"        ,  CVL },
    { "CVS"        ,  CVS },
    { "DATE$"      ,  DATE_S },
    { "DECLARE"    ,  DECLARE },
    { "DEF"        ,  DEF },
    { "DIM"        ,  DIM },
    { "DO"         ,  DO },
    { "DOUBLE"     ,  DOUBLE },
    { "DRAW"       ,  DRAW },
    { "ELSE"       ,  ELSE },
    { "ELSEIF"     ,  ELSEIF },
    { "END"        ,  END },
    { "ENVIRON$"   ,  ENVIRON_S /* also takes strings */ },
    { "EQV"        ,  EQV },
    { "ERASE"      ,  ERASE },
    { "ERL"        ,  ERL },
    { "ERR"        ,  ERR },
    { "ERROR"      ,  ERROR },
    { "EXIT"       ,  EXIT },
    { "EXP"        ,  EXP },
    { "FIELD"      ,  FIELD },
    { "FILEATTR"   ,  FILEATTR /* two args */ },
    { "FILES"      ,  FILES },
    { "FIX"        ,  FIX },
    { "FOR"        ,  FOR },
    { "FRE"        ,  FRE       /* also takes strings */ },
    { "FREEFILE"   ,  FREEFILE },
    { "FUNCTION"   ,  FUNCTION },
    { "GET"        ,  GET },
    { "GOSUB"      ,  GOSUB },
    { "GOTO"       ,  GOTO },
    { "HEX$"       ,  HEX_S },
    { "IF"         ,  IF },
    { "IMP"        ,  IMP },
    { "INKEY$"     ,  INKEY_S },
    { "INP"        ,  INP },
    { "INPUT"      ,  INPUT },
    { "INPUT$"     ,  INPUT_S },
    { "INSTR"      ,  INSTR },
    { "INT"        ,  INT },
    { "INTEGER"    ,  INTEGER },
    { "IS"         ,  IS },
    { "KEY"        ,  KEY },
    { "KILL"       ,  KILL },
    { "LBOUND"     ,  LBOUND /* array */ },
    { "LCASE$"     ,  LCASE_S },
    { "LEFT$"      ,  LEFT_S },
    { "LEN"        ,  LEN },
    { "LET"        ,  LET },
    { "LINE"       ,  LINE },
    { "LIST"       ,  LIST },
    { "LOC"        ,  LOC },
    { "LOCATE"     ,  LOCATE },
    { "LOCK"       ,  LOCK },
    { "LOF"        ,  LOF },
    { "LOG"        ,  LOG },
    { "LONG"       ,  LONG },
    { "LOOP"       ,  LOOP },
    { "LPOS"       ,  LPOS },
    { "LPRINT"     ,  LPRINT },
    { "LSET"       ,  LSET },
    { "LTRIM$"     ,  LTRIM_S },
    { "MID$"       ,  MID_S },
    { "MKD$"       ,  MKD_S },
    { "MKDIR"      ,  MKDIR },
    { "MKI$"       ,  MKI_S },
    { "MKL$"       ,  MKL_S },
    { "MKS$"       ,  MKS_S },
    { "MOD"        ,  MOD },
    { "NEXT"       ,  NEXT },
    { "NOISE"      ,  NOISE },
    { "NOT"        ,  NOT },
    { "OCT$"       ,  OCT_S },
    { "OFF"        ,  OFF },
    { "ON"         ,  ON },
    { "OPEN"       ,  OPEN },
    { "OPTION"     ,  OPTION },
    { "OR"         ,  OR },
    { "OUT"        ,  OUT },
    { "OUTPUT"     ,  OUTPUT },
    { "PAINT"      ,  PAINT },
    { "PALETTE"    ,  PALETTE },
    { "PEEK"       ,  PEEK },
    { "PLAY"       ,  PLAY },
    { "PMAP"       ,  PMAP },
    { "POINT"      ,  POINT /* gfx */ },
    { "POKE"       ,  POKE },
    { "POS"        ,  POS },
    { "PRESET"     ,  PRESET },
    { "PRINT"      ,  PRINT },
    { "PSET"       ,  PSET },
    { "PUT"        ,  PUT },
    { "RANDOM"     ,  RANDOM },
    { "RANDOMIZE"  ,  RANDOMIZE },
    { "READ"       ,  READ },
    { "REDIM"      ,  REDIM },
    { "RESTORE"    ,  RESTORE },
    { "RESUME"     ,  RESUME },
    { "RETURN"     ,  RETURN },
    { "RIGHT$"     ,  RIGHT_S },
    { "RMDIR"      ,  RMDIR },
    { "RND"        ,  RND },
    { "RSET"       ,  RSET },
    { "RTRIM$"     ,  RTRIM_S },
    { "SCREEN"     ,  SCREEN },
    { "SEEK"       ,  SEEK },
    { "SEG"        ,  SEG },
    { "SELECT"     ,  SELECT },
    { "SGN"        ,  SGN },
    { "SHARED"     ,  SHARED },
    { "SHELL"      ,  SHELL },
    { "SIN"        ,  SIN },
    { "SINGLE"     ,  SINGLE },
    { "SLEEP"      ,  SLEEP },
    { "SOUND"      ,  SOUND },
    { "SPACE$"     ,  SPACE_S },
    { "SPC"        ,  SPC },
    { "SQR"        ,  SQR },
    { "STATIC"     ,  STATIC },
    { "STEP"       ,  STEP },
    { "STOP"       ,  STOP },
    { "STR$"       ,  STR_S },
    { "STRING"     ,  STRING },
    { "STRING$"    ,  STRING_S },
    { "SUB"        ,  SUB },
    { "SWAP"       ,  SWAP },
    { "SYSTEM"     ,  SYSTEM },
    { "TAB"        ,  TAB },
    { "TAN"        ,  TAN },
    { "THEN"       ,  THEN },
    { "TIME$"      ,  TIME_S },
    { "TIMER"      ,  TIMER },
    { "TO"         ,  TO },
    { "TYPE"       ,  TYPE },
    { "UBOUND"     ,  UBOUND /* array */ },
    { "UCASE$"     ,  UCASE_S },
    { "UNTIL"      ,  UNTIL },
    { "USING"      ,  USING },
    { "VAL"        ,  VAL },
    { "VARPTR"     ,  VARPTR   /* any */ },
    { "VARPTR$"    ,  VARPTR_S /* any */ },
    { "VARSEG"     ,  VARSEG   /* any */ },
    { "VIEW"       ,  VIEW },
    { "WAIT"       ,  WAIT },
    { "WEND"       ,  WEND },
    { "WHILE"      ,  WHILE },
    { "WIDTH"      ,  WIDTH },
    { "WINDOW"     ,  WINDOW },
    { "WRITE"      ,  WRITE },
    { "XOR"        ,  XOR },
    { "_BIN$"      ,  qbcBIN_S },
    { "_BIT"       ,  qb64BIT },
    { "_BYTE"      ,  qb64BYTE },
    { "_DISPLAY"   ,  qb64DISPLAY },
    { "_DONTBLEND" ,  qb64DONTBLEND },
    { "_KEYDOWN"   ,  qb64KEYDOWN },
    { "_NEWIMAGE"  ,  qb64NEWIMAGE },
    { "_SNDBAL"    ,  qb64SNDBAL },
    { "_SNDCLOSE"  ,  qb64SNDCLOSE },
    { "_SNDCOPY"   ,  qb64SNDCOPY },
    { "_SNDOPEN"   ,  qb64SNDOPEN },
    { "_SNDPAUSE"  ,  qb64SNDPAUSE },
    { "_SNDPAUSED" ,  qb64SNDPAUSED },
    { "_SNDPLAYCOPY",  qb64SNDPLAYCOPY },
    { "_SNDPLAYING",  qb64SNDPLAYING },
    { "_SNDPLAY"   ,  qb64SNDPLAY },
    { "_SNDRATE"   ,  qb64SNDRATE },
    { "_SNDRAWDONE",  qb64SNDRAWDONE },
    { "_SNDRAWLEN" ,  qb64SNDRAWLEN },
    { "_SNDRAWOPEN",  qb64SNDRAWOPEN },
    { "_SNDRAW"    ,  qb64SNDRAW },
    { "_SNDSTOP"   ,  qb64SNDSTOP },
    { "_SHL"       ,  qbcSHL },
    { "_SHR"       ,  qbcSHR },
    { "_ASIN"      ,  qbcASIN },
    { "_ACOS"      ,  qbcACOS },
    { "_SINH"      ,  qbcSINH },
    { "_COSH"      ,  qbcCOSH },
    { "_TANH"      ,  qbcTANH },
    { "_ASINH"     ,  qbcASINH },
    { "_ACOSH"     ,  qbcACOSH },
    { "_ATANH"     ,  qbcATANH },
    { "_ATAN2"     ,  qbcATAN2 },
    { "_UNSIGNED"  ,  qb64UNSIGNED }
};
static const struct
{
    QBtoken token1, token2;
    QBtoken result;
} ComboTokens_Const[] =
{
    { DEF,SEG,       DEFSEG },
    { END,DEF,       ENDDEF },
    { END,FUNCTION,  ENDFUNCTION },
    { END,IF,        ENDIF },
    { END,SELECT,    ENDSELECT },
    { END,SUB,       ENDSUB },
    { END,TYPE,      ENDTYPE },
    { EXIT,DEF,      EXITDEF },
    { EXIT,DO,       EXITDO },
    { EXIT,FOR,      EXITFOR },
    { EXIT,FUNCTION, EXITFUNCTION },
    { EXIT,SUB,      EXITSUB },
    { LINE,INPUT,    LINE_INPUT },
    { ON,ERROR,      ON_ERROR },
    { ON,KEY,        ON_KEY },
    { ON,PLAY,       ON_PLAY },
    { ON,TIMER,      ON_TIMER },
    { PALETTE,USING, PALETTE_USING },
    { RESUME,NEXT,   RESUME_NEXT },
    { VIEW,PRINT,    VIEW_PRINT },
    { WIDTH,HASHSYM, WIDTH_F },
    { WIDTH,LPRINT,  WIDTH_LPRINT },
    { qb64UNSIGNED, INTEGER,  qb64UNSIGNEDINTEGER },
    { qb64UNSIGNED, LONG,     qb64UNSIGNEDLONG },
    { qb64UNSIGNED, qb64BIT,  qb64UNSIGNEDBIT },
    { qb64UNSIGNED, qb64BYTE, qb64UNSIGNEDBYTE },
};

struct saved_yy_context
{
    LexLocation loc;
    LexLocation loc_at_begin_of_token;
    std::FILE *fp;

    std::vector<char>::iterator buffer_limit;
    std::vector<char>::iterator buffer_cursor;
    std::vector<char>::iterator buffer_marker;
    std::vector<char>::iterator buffer_begin;
    std::vector<char> buffer;
    int condition;

    void InitBuffer()
    {
        buffer.clear();
        buffer_limit  = buffer.begin();
        buffer_cursor = buffer.begin();
        buffer_marker = buffer.begin();
        buffer_begin  = buffer.begin();
    }
    void Catchup()
    {
        buffer.erase(buffer.begin(), buffer_cursor);
        buffer_cursor = buffer.begin();
        buffer_marker = buffer.begin();
        buffer_begin  = buffer.begin();
        buffer_limit  = buffer.end();
        loc_at_begin_of_token = loc;
        FillBuffer(1);
    }
    void FillBuffer(unsigned n)
    {
        if( (buffer_limit-buffer_cursor) <= n)
            n -= (buffer_limit - buffer_cursor);
        if(!fp || !n) return;

        std::size_t cursor = buffer_cursor - buffer.begin();
        std::size_t marker = buffer_marker - buffer.begin();
        std::size_t begin  = buffer_begin  - buffer.begin();

        for(; n > 0; --n)
        {
            int c;
            do c = std::fgetc(fp); while(c == '\r' || c == '\32');
            if(c == '_')
            {
                do { c = std::fgetc(fp); } while(c == '\r' || c == '\32');
                if(c == '\n') c = std::fgetc(fp);
                else { ungetc(c, fp); c = '_'; }
            }
            if(c == EOF) { std::fclose(fp); fp = nullptr; break; }
            switch(c)
            {
                case '\n': ++loc.line; loc.col=1; break;
                case '\t': loc.col = 1 + ((loc.col+7)&~7); break;
                default: ++loc.col;
            }

            buffer.push_back( char(c) );
        }
        if(!fp)
        {
            buffer.push_back('\32');
        }
        buffer_cursor = buffer.begin() + cursor;
        buffer_marker = buffer.begin() + marker;
        buffer_begin  = buffer.begin() + begin;
        buffer_limit = buffer.end();
    }

    void Debug(int state, char ch)
    {
#ifdef LEXER_DEBUG
        dprintf(dp_lexer, "Lexer state%6d with '%c'; Row %d,column %d - Buffer(%4u/%2u): %s\n",
                state,ch, loc_at_begin_of_token.line,loc_at_begin_of_token.col,
                (unsigned)buffer.size(), (unsigned) Get().size(), Get().c_str());
#else
        state=state; ch=ch;
#endif
    }

    std::string Get(std::size_t skip_begin=0, std::size_t skip_end=0) const
    {
        return std::string( buffer_begin+skip_begin, buffer_cursor-skip_end );
    }
};

static saved_yy_context ctx;
static std::list<saved_yy_context> contexts;
static QBtoken prev_return = CRLF;
static std::unordered_map<std::string, QBtoken> ReservedWords;
static std::unordered_map<unsigned, std::unordered_map<unsigned, QBtoken>> ComboTokens;

LexerResult yylval;

void yyStartFile(std::FILE *fp, const char *name)
{
    ctx.loc.line     = 1;
    ctx.loc.col      = 1;
    ctx.loc.filename = name;
    /* .fp and .state are filled at $INCLUDE. */
    ctx.fp           = fp;
    ctx.condition    = -1;
    ctx.InitBuffer();

    if(ReservedWords.empty())
    {
        for(const auto& w: ReservedWords_Const)
            ReservedWords.insert( { w.string, w.token } );
        for(const auto& c: ComboTokens_Const)
            ComboTokens[c.token1].insert( { c.token2, c.result } );
    }

    SetDefVarType("ABCDEFGHIJKLMNOPQRSTUVWXYZ", "!"); // DEFSNG A-Z
}
bool yyEndFile(void)
{
    if(ctx.fp) std::fclose(ctx.fp);
    if(contexts.empty()) return false;

    ctx = std::move( contexts.back() );
    contexts.pop_back();
    return true;
}
const LexLocation& yyGetLocation()
{
     return ctx.loc_at_begin_of_token;
}

const std::string LexLocation::stringify() const
{
    return filename + ":" + std::to_string(line) + ":" + std::to_string(col);
}

void yyerror(const std::string& msg)
{
    yyerror(msg, ctx.loc);
}
void yyerror(const std::string& msg, const LexLocation& loc)
{
    dprintf(dp_errors, "\n%s: Error: %s\n", loc.stringify().c_str(), msg.c_str());
}

QBtoken LexerGetToken()
{
/*!re2c

digit = [0-9];
bindigit = [0-1];
octdigit = [0-7];
hexdigit = [0-9A-Fa-f];
blank = [ \t]+;
sign  = [-+];

/* Newlines also yank spaces, so we can get more relevant lex-locations for statements */
newlines = "\n" [ \t\n]* ;

/* Colon characters do the same. */
colons   = ":"  [: \t]*  ;

typelengthsuffix  = ([1-9] [0-9]*)?;
stringtypesuffix  = "$" typelengthsuffix?;
floattypesuffix   = [!#];
unsignedtypesuffix = "~";
bittypesuffix     = unsignedtypesuffix? "`" typelengthsuffix?;
integertypesuffix = unsignedtypesuffix? ( "%" | "%%" | "&" ) | bittypesuffix;
typesuffix = stringtypesuffix | integertypesuffix | floattypesuffix;

re2c:define:YYCTYPE = "unsigned char";
re2c:yych:conversion = 1;
re2c:indent:string = "    ";
re2c:define:YYDEBUG  = "ctx.Debug";
*/
    enum YYCONDTYPE { yycINITIAL, yycDEFCMD, yycDATA, yycCOMMENT, yycCOMMENT_NOTMETA, yycINCLUDE };

    #define YYGETCONDITION()   ctx.condition
    #define YYSETCONDITION(n)  ctx.condition=(n)
    #define YYLIMIT   ctx.buffer_limit
    #define YYCURSOR  ctx.buffer_cursor
    #define YYMARKER  ctx.buffer_marker
    #define YYFILL(n) ctx.FillBuffer(n)

    static std::pair<LexerResult,QBtoken> pending;
    static bool token_pending = false;

    #define Return(val) do { prev_return = (QBtoken)(val); goto CheckReturn; } while(0)

    if(token_pending)
    {
        yylval = std::move( pending.first );
        token_pending = false;
        return pending.second;
    }

//RESCAN:
    ctx.Catchup();
/*!re2c

    /* Grammar begins: */

    /* Exception: DEFTYPE commands */
<INITIAL>'DEFINT'  => DEFCMD   { yylval.numtype = "%"; ctx.Catchup(); yylval.str.clear(); goto yyc_DEFCMD; }
<INITIAL>'DEFSNG'  => DEFCMD   { yylval.numtype = "!"; ctx.Catchup(); yylval.str.clear(); goto yyc_DEFCMD; }
<INITIAL>'DEFDBL'  => DEFCMD   { yylval.numtype = "#"; ctx.Catchup(); yylval.str.clear(); goto yyc_DEFCMD; }
<INITIAL>'DEFSTR'  => DEFCMD   { yylval.numtype = "$"; ctx.Catchup(); yylval.str.clear(); goto yyc_DEFCMD; }
<INITIAL>'DEFLNG'  => DEFCMD   { yylval.numtype = "&"; ctx.Catchup(); yylval.str.clear(); goto yyc_DEFCMD; }
<INITIAL>'_DEFINE' => DEFCMD   { yylval.numtype.clear(); ctx.Catchup(); yylval.str.clear(); goto yyc_DEFCMD; }

    /* Exception: DATA */
<INITIAL>'DATA' blank? => DATA  { Return(DATA); }

    /* Exception: Comments */
<INITIAL>'REM'     => COMMENT  {
                      #ifdef LEXER_DEBUG
                        dprintf(dp_lexer, "REM acquired. Prev_return = %s\n", GetTokenName(prev_return).c_str());
                      #endif
                        if(prev_return == THEN || prev_return == ELSE)
                        {
                            // If REM follows a THEN or ELSE, it is treated as a statement.
                            // This means that "IF 0 THEN REM xx" is a shortif, not a longif.
                            Return(REM);
                        }
                        if(prev_return != COLONCHAR
                        && prev_return != CRLF
                        && prev_return != NUMBER)
                        {
                            dprintf(dp_errors, "Error: Unexpected REM\n");
                            YYSETCONDITION(yycINITIAL);
                            Return(REM);
                        }
                        /* REM starts a comment only when at statement boundary */
                        /* REM always ends in a CRLF, so return the CRLF now.   */
                        return CRLF;
                      }
<INITIAL>"'"       => COMMENT { return(CRLF); }

<INITIAL>     [a-zA-Z] [a-z_A-Z0-9]* typesuffix? |
              "_"      [a-z_A-Z0-9]+ typesuffix?

    { yylval.str = std::move(ctx.Get());

      auto i = ReservedWords.find(UCase(yylval.str));
      if(i != ReservedWords.end())
      {
          Return(i->second);
      }

      if( UCase(yylval.str.substr(0,2)) == "FN" )
          Return(FNIDENTIFIER);

      Return(IDENTIFIER);
    }

<INITIAL> sign? digit+ integertypesuffix {
    //fprintf(stderr, "Integer: '%s'\n", yytext);
    auto s = ctx.Get(); const char* yytext = s.c_str();
    char *endp;
    yylval.had_sign = *yytext=='-' || *yytext=='+';
    if(*yytext=='+') ++yytext;

    yylval.number  = (strtoll(yytext,&endp,10));
    yylval.numtype = endp;
    Return(NUMBER);
  }

<INITIAL>"+"? '&H' hexdigit+ integertypesuffix? {
    //fprintf(stderr, "Hex integer: '%s'\n", yytext);
    auto s = ctx.Get(); const char* yytext = s.c_str();
    yylval.had_sign = *yytext=='+';
    if(*yytext=='+') ++yytext;
    char*endp;
    yylval.number  = (strtoll(yytext+2,&endp,16));
    yylval.numtype = *endp ? endp : "%";
    //fprintf(stderr, "-> %Lg\n", yylval.number);
    Return(NUMBER);
  }

<INITIAL>"+"? '&O' octdigit+ integertypesuffix? {
    //fprintf(stderr, "Octal integer: '%s'\n", yytext);
    auto s = ctx.Get(); const char* yytext = s.c_str();
    yylval.had_sign = *yytext=='+';
    if(*yytext=='+') ++yytext;
    char*endp;
    yylval.number  = (strtoll(yytext+2,&endp,8));
    yylval.numtype = *endp ? endp : "%";
    //fprintf(stderr, "-> %Lg\n", yylval.number);
    Return(NUMBER);
  }

<INITIAL>"+"? '&B' bindigit+ integertypesuffix? {
    //fprintf(stderr, "Binary integer: '%s'\n", yytext);
    auto s = ctx.Get(); const char* yytext = s.c_str();
    yylval.had_sign = *yytext=='+';
    if(*yytext=='+') ++yytext;
    char*endp;
    yylval.number  = (strtoll(yytext+2,&endp,2));
    yylval.numtype = *endp ? endp : "%";
    //fprintf(stderr, "-> %Lg\n", yylval.number);
    Return(NUMBER);
  }

<INITIAL> sign? digit+ ("." digit*)? ('e' sign? digit+)? floattypesuffix?   |
          sign?         "." digit+   ('e' sign? digit+)? floattypesuffix? {
    auto s = ctx.Get(); const char* yytext = s.c_str();
    char*endp;
    yylval.had_sign = *yytext=='-' || *yytext=='+';
    if(*yytext=='+') ++yytext;
    yylval.number   = (strtold(yytext,&endp));
    yylval.numtype  = endp;
    if(!*endp)
    {
        yylval.numtype = "#";
        if(yylval.number == (float)yylval.number) yylval.numtype = "!";
        if(yylval.number == (long)yylval.number)  yylval.numtype = "&";
        if(yylval.number == (short)yylval.number) yylval.numtype = "%";
    }
    Return(NUMBER);
  }

    /* GO TO is recognized, but GO must not be a token. */
<INITIAL>'GO' blank 'TO'  { Return(GOTO); }
<INITIAL>"?"              { Return(PRINT); }

<INITIAL> colons          { Return(COLONCHAR); }
<INITIAL> newlines        { Return(CRLF); }

<INITIAL> "#"   blank? { Return (HASHSYM); }
<INITIAL> "<="  blank? { Return (LESSOREQUAL); }
<INITIAL> ">="  blank? { Return (GREATEROREQUAL); }
<INITIAL> "<>"  blank? |
          "><"  blank? |
          "!="  blank? { Return (NOTEQUAL); }

<INITIAL>'"' [^"\n]* ["\n]  {
    yylval.str = ctx.Get(1,1); // Skip '#' and remove '"'
    //fprintf(stderr, "String: '%s'\n", yylval.str.c_str());
    if(YYCURSOR[-1] == '\n') --YYCURSOR; // unput linefeed
    Return (STRINGCONST);
  }

<INITIAL>"-"  blank? { Return (MINUS); }
<INITIAL>"<"  blank? { Return (LESSTHAN); }
<INITIAL>">"  blank? { Return (GREATERTHAN); }
<INITIAL>"="  blank? { Return (EQUALS); }
<INITIAL>"+"  blank? { Return (PLUS); }
<INITIAL>"*"  blank? { Return (ASTERISK); }
<INITIAL>"/"  blank? { Return (SLASH); }
<INITIAL>"\\" blank? { Return (BKSLASH); }
<INITIAL>","  blank? { Return (COMMA); }
<INITIAL>";"  blank? { Return (SEMICOLON); }
<INITIAL>"("  blank? { Return (OPENPARENS); }
<INITIAL>")"  blank? { Return (CLOSEPARENS); }
<INITIAL>"."  blank? { Return (PERIOD); }
<INITIAL>"^"  blank? { Return (CIRCUMFLEX); }

<INITIAL>blank  { /* blank   */ ctx.Catchup(); goto yyc_INITIAL; }

<DEFCMD>  [a-zA-Z] "-" [a-zA-Z] { char first = std::toupper(ctx.buffer_begin[0]);
                                  char last  = std::toupper(ctx.buffer_begin[2]);
                                  if(first > last) std::swap(first, last);
                                  for(char c=first; c<=last; ++c) yylval.str += c;
                                  ctx.Catchup(); goto yyc_DEFCMD;
                                }
<DEFCMD>  [a-zA-Z][, \t\n':;]   { yylval.str += std::toupper(ctx.buffer_begin[0]);
                                  --YYCURSOR; // Discard that trailing character
                                  ctx.Catchup(); goto yyc_DEFCMD;
                                }
<DEFCMD>  [, \t]+               { /* Ignore comma and space */ ctx.Catchup();; goto yyc_DEFCMD; }
<DEFCMD>  .        => INITIAL   { YYCURSOR = ctx.buffer_begin; Return (qb64DEFINE); /* end of def */ }

<DATA>  '"' [^"\n]* ["\n]  {
    yylval.str = ctx.Get(1,1); // Skip '#' and remove '"'
    //fprintf(stderr, "Data string (quoted): '%s'\n", yylval.str.c_str());
    if(YYCURSOR[-1] == '\n') --YYCURSOR; // unput linefeed
    Return (STRINGCONST);
  }

<DATA>  [ \t]+      { /* Ignore space in DATA */ ctx.Catchup(); goto yyc_DATA; }
<DATA>  [^":,\n \t][^:,\n]*  {
    yylval.str = Trim(ctx.Get(0,0));
    //fprintf(stderr, "Data string (unquoted): '%s'\n", yylval.str.c_str());
    /* ltrim and rtrim the string - qb does that too. */
    Return (STRINGCONST);
  }
<DATA>  "," blank?            { Return (COMMA); }
<DATA>  .          => INITIAL { YYCURSOR = ctx.buffer_begin; goto yyc_INITIAL; /* end of data */ }

    /* Note that in QuickBASIC, $DYNAMICabcdef is interpreted as $DYNAMIC.
     * We replicate the same behavior here.
     */
<COMMENT>  '$DYNAMIC'       { Return (metaDYNAMIC); }
<COMMENT>  '$STATIC'        { Return (metaSTATIC);  }
<COMMENT>  '$MAGICAL'       {
    dprintf(dp_warnings, "There is no $MAGICAL keyword. It is qbc's term for arrays that are passed as function parameters, because they can be of any type and size, only determined at runtime. Any occurrence of $MAGICAL is treated just as a regular comment.\n");
                              ctx.Catchup();
                              goto yyc_COMMENT;
                            }
<COMMENT>  '$INCLUDE:' blank "'" => INCLUDE { ctx.Catchup(); goto yyc_INCLUDE; }
<COMMENT>  newlines          => INITIAL { Return (CRLF); }
<COMMENT>  blank            { /* ignore spaces */ goto yyc_COMMENT; }
                   /* anything else: meta-statements no longer recognized */
<COMMENT>  .        :=> COMMENT_NOTMETA

<COMMENT_NOTMETA>  newlines  => INITIAL  { Return (CRLF); }
<COMMENT_NOTMETA>  [^\n]+    { /* Ignore longest possible blobs until newline */ goto yyc_COMMENT_NOTMETA; }

<INCLUDE>  [^'\n]* => INITIAL {
    auto s = ctx.Get(); const char* yytext = s.c_str();
    dprintf(dp_lexer, "Meta $INCLUDE: '%s'\n", yytext);

    contexts.emplace_back( std::move(ctx) );

    FILE* fp = std::fopen(yytext, "rt");
    if(!fp)
    {
        dprintf(dp_errors, "Error: unable to open include file %s", yytext);
        ctx.fp = nullptr;
        return static_cast<QBtoken>(0);
    }

    yyStartFile(fp, yytext);
  }
<INCLUDE>  "'"  :=> COMMENT_NOTMETA

<*>      "\032" => INITIAL { // ^Z
    bool e = yyEndFile();
    if(!e) { Return (0); }
    ctx.Catchup(); goto yyc_INITIAL;
}
<*> .      => INITIAL { dprintf(dp_errors, "Unknown character: '%c'\n", YYCURSOR[-1]); goto yyc_INITIAL; }
*/

CheckReturn:
    auto i = ComboTokens.find(prev_return);
    if(i == ComboTokens.end())
        return prev_return;

    LexerResult first_result = std::move(yylval);
    QBtoken first_token  = prev_return;
    QBtoken second_token = LexerGetToken();
    auto j = i->second.find(second_token);
    if(j != i->second.end())
        return j->second;

    // Oops. Wasn't a combo. Now we have two tokens.
    // Let's put the second one pending, and return the first one.
    token_pending = true;
    pending.first  = std::move(yylval);
    pending.second = second_token;

    yylval = std::move(first_result);
    return first_token;
}
