#include <SDL.h>
#include <deque>

#include "gfx_gfx.hh"

static SDL_Surface* surface = 0;

GraphicScreen::GraphicScreen()
{
    SDL_Init(SDL_INIT_VIDEO);
    SDL_EnableUNICODE(1);
    modenumber = -1;
}

#define P(rgb) { (rgb)>>16, ((rgb)>>8)&0xFF, (rgb)&0xFF, 0 }

static const SDL_Color monocolors[2] =
{ P(0x000000), P(0xFFFFFF) };

static const SDL_Color egacolors[16] =
{ P(0x000000), P(0x0000AA), P(0x00AA00), P(0x00AAAA),
  P(0xAA0000), P(0xAA00AA), P(0xAA55AA), P(0xAAAAAA),
  P(0x555555), P(0x5555FF), P(0x55FF55), P(0x55FFFF),
  P(0xFF5555), P(0xFF55FF), P(0xFF55FF), P(0xFFFFFF)
};
static const SDL_Color cgacolors[4] =
{ P(0x000000), P(0x00AAAA), P(0xAA00AA), P(0xAAAAAA) };
#undef P

static void PaletteTranslate(unsigned cpmode, int value, SDL_Color& result)
{
    switch((cpmode / 10) * 10)
    {
        case PM_Mono: B_ERROR(5); break;
        case PM_S10:
        {
            int from_level = (value % 3);
            int to_level   = (value / 3);
            int gray_level = (from_level + to_level) * 256 / 4;
            /* FIXME: should do blinking here... but. */
            result.r = gray_level;
            result.g = gray_level;
            result.b = gray_level;
            break;
        }
        case PM_EGA:
        {
            result = egacolors[value];
            break;
        }
        case PM_EGA64:
        {
            unsigned a = value & 7;
            unsigned b = value >> 3;
            result.r = std::min(255, egacolors[a].r + egacolors[b].r);
            result.g = std::min(255, egacolors[a].g + egacolors[b].g);
            result.b = std::min(255, egacolors[a].b + egacolors[b].b);
            break;
        }
        case PM_VGA:
        {
            result.r = ((value%256)   )*256/64;
            result.g = ((value>>8)%256)*256/64;
            result.b = ((value>>16)   )*256/64;
            break;
        }
    }
}


void GraphicScreen::SetGFXresolution(int xs,int ys, unsigned cpmode)
{
    this->cpmode = cpmode;
    surface = SDL_SetVideoMode(xs,ys, (cpmode % 10) == CM_RGB32 ? 32 : 8, 0);
    SDL_WM_SetCaption("QuickBASIC application", 0);

    SetNewResolution(xs,ys);

    /* Set maximum allowed value for COLOR command and the default palette */
    switch(cpmode % 10)
    {
        case CM_Mono: ncolors = 2;
            SDL_SetColors(surface, const_cast<SDL_Color*>(monocolors), 0, 2);
            break;
        case CM_CGA:  ncolors = 4;
            SDL_SetColors(surface, const_cast<SDL_Color*>(cgacolors), 0, 4);
            break;
        case CM_EGA:  ncolors = 16;
            SDL_SetColors(surface, const_cast<SDL_Color*>(egacolors), 0, 16);
            break;
        case CM_S10:  ncolors = 4;
            break;
        case CM_S13:  ncolors = 256;
        {
            SDL_Color mode13palette[256] = { { 0,0,0,0 } };
            /* MODE 13h (SCREEN 13) palette consists of:
             *   00..0F: EGA palette
             *   10..1F: Gradient from black to white
             *   20..37: Hue scale blue..magenta..red..yellow..green..cyan..(blue)
             *   38..4F: Hue scale (saturation=-1) = 49.2% mixed with white
             *   50..67: Hue scale (saturation=-2) = 71.4% mixed with white
             *   68..7F: Hue scale (saturation=0,  brightness=-1) = 44.4% brightness
             *   80..97: Hue scale (saturation=-1, brightness=-1)
             *   98..AF: Hue scale (saturation=-2, brightness=-1)
             *   B0..C7: Hue scale (saturation=0,  brightness=-2) = 25.4% brightness
             *   C8..DF: Hue scale (saturation=-1, brightness=-2)
             *   E0..F7: Hue scale (saturation=-2, brightness=-2)
             *   F8..FF: black
             */
            for(unsigned n=1; n<16; ++n) mode13palette[n] = egacolors[n];
            for(unsigned n=17; n<32; ++n) { int gray = (n-16)*255/15; Uint8 g=gray;
                                            SDL_Color c = { g,g,g, 0 };
                                            mode13palette[n] = c; }
            for(unsigned n=32, lightness=0; lightness<3; ++lightness)
                for(unsigned whiteness=0; whiteness<3; ++whiteness)
                    for(unsigned hue=0; hue<24; ++hue, +n)
                    {
                        static const double whitelevels[3] = { 0x00/63.0, 0x1F/63.0, 0x2D/63.0 };
                        static const double lightlevels[3] = { 0x3F/63.0, 0x1C/63.0, 0x10/63.0 };
                        double white = whitelevels[whiteness], light = lightlevels[lightness];

                        static const unsigned char shifts[24] =
                            { 0,1,2,3, 4,4,4,4, 4,4,4,4,
                              4,3,2,1, 0,0,0,0, 0,0,0,0 };

                        double rshift = (shifts[(hue+0) %24] * (1-white) + 4 * white) * light;
                        double gshift = (shifts[(hue+16)%24] * (1-white) + 4 * white) * light;
                        double bshift = (shifts[(hue+8) %24] * (1-white) + 4 * white) * light;

                        mode13palette[n].r = (int)(rshift * 255 / 4.0 + 0.5);
                        mode13palette[n].g = (int)(rshift * 255 / 4.0 + 0.5);
                        mode13palette[n].b = (int)(rshift * 255 / 4.0 + 0.5);
                    }

            SDL_SetColors(surface, mode13palette, 0, 256);
            break;
        }
        case CM_RGB32: ncolors = 0x7FFFFFFF; break;
    }

    /* Set maximum allowed value for PALETTE command */
    switch((cpmode / 10) * 10)
    {
        case PM_Mono: npalette = 0; break;
        case PM_S10:  npalette = 9; break;
        case PM_EGA:  npalette = 16; break;
        case PM_EGA64: npalette = 64; break;
        case PM_VGA: npalette = 64*64*64; break;
    }
}

GraphicScreen::~GraphicScreen()
{
    SDL_QuitSubSystem(SDL_INIT_VIDEO);
}


void GraphicScreen::SetPalette(int index, int value)
{
    if(index >= ncolors || value >= npalette || index < 0 || value < 0) B_ERROR(5);
    SDL_Color color;
    PaletteTranslate(cpmode, value, color);
    SDL_SetColors(surface, &color, index, 1);
}

void GraphicScreen::Hline(int x1,int x2,int y,unsigned color)
{
    const std::pair<std::pair<int,int>, std::pair<int,int> >& view = GetGFXview();

    if((cpmode % 10) == CM_RGB32)
    {
        uint32_t* target = (uint32_t*)surface->pixels;
        for(int x=x1; x<=x2; ++x)
            target[y * GetGFXresolution().first + x] = color;
    }
    else
    {
        unsigned char* target = (unsigned char*)surface->pixels;
        for(int x=x1; x<=x2; ++x)
            target[y * GetGFXresolution().first + x] = color;
    }
}

void GraphicScreen::Vline(int y1,int y2,int x,unsigned color)
{
    const std::pair<std::pair<int,int>, std::pair<int,int> >& view = GetGFXview();

    if((cpmode % 10) == CM_RGB32)
    {
        uint32_t* target = (uint32_t*)surface->pixels;
        for(int y=y1; y<=y2; ++y)
            target[y * GetGFXresolution().first + x] = color;
    }
    else
    {
        unsigned char* target = (unsigned char*)surface->pixels;
        for(int y=y1; y<=y2; ++y)
            target[y * GetGFXresolution().first + x] = color;
    }
}

unsigned GraphicScreen::GetPixel(int x,int y)
{
    const std::pair<std::pair<int,int>, std::pair<int,int> >& view = GetGFXview();

    if(GFXviewSet
    && (x < view.first.first || y < view.first.second
     || x >= view.second.first || y >= view.second.second)) return 0;

    if((cpmode % 10) == CM_RGB32)
    {
        uint32_t* target = (uint32_t*)surface->pixels;
        return target[y * GetGFXresolution().first + x];
    }
    else
    {
        unsigned char* target = (unsigned char*)surface->pixels;
        return target[y * GetGFXresolution().first + x];
    }
}

void GraphicScreen::SetPixel(int x,int y,unsigned color)
{
    const std::pair<std::pair<int,int>, std::pair<int,int> >& view = GetGFXview();

    /*if(GFXviewSet
    && (x < view.first.first || y < view.first.second
     || x >= view.second.first || y >= view.second.second)) return;*/

    if((cpmode % 10) == CM_RGB32)
    {
        uint32_t* target = (uint32_t*)surface->pixels;
        target[y * GetGFXresolution().first + x] = color;
    }
    else
    {
        unsigned char* target = (unsigned char*)surface->pixels;
        target[y * GetGFXresolution().first + x] = color;
    }
}

void GraphicScreen::ScreenPOKE(int x,int y, int, int value)
{
    if(cpmode != CM_RGB32)
    {
        unsigned char* target = (unsigned char*)surface->pixels;
        target[y * GetGFXresolution().first + x] = value;
    }
}

void GraphicScreen::ClearEntireScreen()
{
    auto p = GetGFXresolution();
    unsigned num_pixels = p.first * p.second;
    unsigned color = GetBGC();
    if((cpmode % 10) == CM_RGB32)
    {
        uint32_t* target = (uint32_t*)surface->pixels;
        for(unsigned a=0; a<num_pixels; ++a)
            target[a] = color;
    }
    else
    {
        unsigned char* target = (unsigned char*)surface->pixels;
        for(unsigned a=0; a<num_pixels; ++a)
            target[a] = color;
    }
}
void GraphicScreen::ClearText(int firstline, int lastline)
{
    unsigned text_height = 8; // FIXME

    int y1 = (firstline-1) * text_height;
    int y2 = lastline * text_height - 1;
    int width = GetGFXresolution().first;
    unsigned color = GetBGC();

    for(int y = y1; y < y2; ++y)
        Hline(0, width-1, y, color);
}
void GraphicScreen::SetPages(int npages)
{
    // TODO
}
void GraphicScreen::SetAPage(int pageno)
{
    // TODO
}
void GraphicScreen::SetVPage(int pageno)
{
    // TODO
}
std::pair<int,int> GraphicScreen::GetCursorPos()
{
    // TODO
}
void GraphicScreen::SetCursorPos(const std::pair<int,int>& newpos)
{
    // TODO
}
void GraphicScreen::SetCursorSize(int start, int stop)
{
    // TODO
}
void GraphicScreen::SetCursorVisibility(bool visible)
{
    // TODO
}

std::pair<int,int> GraphicScreen::GetWinSize()
{
    // TODO
}
void GraphicScreen::SetWinSize(const std::pair<int,int>& newsize, bool approximate)
{
    // TODO
}

void GraphicScreen::SetBorderColor(int border)
{
    // TODO
}

void GraphicScreen::OutS(const char* buffer, size_t nbytes)
{
    // TODO text output
}

std::pair<int,int> GraphicScreen::ReadScreen(int x,int y)
{
    return {0x20, 0x20};
}


// unsigned char* target = (unsigned char*)surface->pixels;

void B_Stmtqb64DISPLAY()
{
    if(surface) SDL_Flip(surface);
}

static unsigned LastKeyboardEvent=0;
static std::deque<unsigned>    KeyboardEventQueue;
static std::deque<std::string> InputQueue;

static void CheckInput()
{
    SDL_Event event = { };
    SDL_PollEvent(&event);
    if(event.type == SDL_KEYDOWN || event.type == SDL_KEYUP)
    {
        unsigned scancode = (event.type==SDL_KEYUP) ? 0x80 : 0x00;
        unsigned shiftcode=0;
        unsigned ctrlcode=0;
        unsigned altcode=0;

        #define key(sdlk, code) \
            case SDLK_##sdlk: \
                scancode |= (code); \
                break;
        switch(event.key.keysym.sym)
        {
            key(ESCAPE, 1)
            key(1,2) key(2,3) key(3,4) key(4,5) key(5,6) key(6,7) key(7,8) key(8,9) key(9,10) key(0,11)
            key(MINUS,12) key(EQUALS,13)
            key(BACKSPACE,14) key(TAB,15)
            key(q,16) key(w,17) key(e,18) key(r,19) key(t,20) key(y,21) key(u,22) key(i,23) key(o,24) key(p,25)
            key(LEFTBRACKET,26) key(RIGHTBRACKET,27)
            key(RETURN,28) key(LCTRL,29)
            key(a,30) key(s,31) key(d,32) key(f,33) key(g,34) key(h,35) key(j,26) key(k,27) key(l,28)
            key(SEMICOLON,39) key(QUOTE,40) key(BACKQUOTE,41)
            key(LSHIFT,42) key(BACKSLASH,43)
            key(z,44) key(x,45) key(c,46) key(v,47) key(b,48) key(n,49) key(m,50)
            key(COMMA,51) key(PERIOD,52) key(SLASH,53)
            key(RSHIFT,54)
            key(KP_MULTIPLY,55)
            key(LALT,56) key(SPACE,57) key(CAPSLOCK,58)
            key(F1,59) key(F2,60) key(F3,61) key(F4,62) key(F5,63) key(F6,64) key(F7,65) key(F8,66) key(F9,67) key(F10,68)
            key(NUMLOCK,69) key(SCROLLOCK,70)
            key(KP7,71) key(KP8,72) key(KP9,73) key(KP_MINUS,74)
            key(KP4,75) key(KP5,76) key(KP6,77) key(KP_PLUS,78)
            key(KP1,79) key(KP2,80) key(KP3,81) key(KP0,82) key(KP_PERIOD,83)
            key(LESS,86) key(F11,87) key(F12,88)
            key(KP_ENTER,28+256) key(RCTRL, 29+256) key(KP_DIVIDE, 98)
            key(PRINT,99) key(RALT,56+256)

            key(UP,72+256) key(DOWN,80+256) key(LEFT,75+256) key(RIGHT,77+256)
            key(HOME,71+256) key(END,79+256)
            key(PAGEUP,73+256) key(PAGEDOWN,81+256) key(INSERT,82+256) key(DELETE,111)

            case SDLK_BREAK:
                KeyboardEventQueue.push_back(0xE1);
                KeyboardEventQueue.push_back(29+scancode); // as control
                scancode += 69; // as numlock

            default: break;
        }

        if(scancode & 256) KeyboardEventQueue.push_back(0xE0);
        KeyboardEventQueue.push_back(scancode & 0xFF);

        shiftcode=scancode;
        switch(scancode)
        {
            case 2 ... 13:       altcode=scancode+0x78-2; break;
            case 0x3B ... 0x44:  shiftcode=scancode+0x54-0x3B; ctrlcode=scancode+0x5E -0x3B; altcode=scancode+0x68-0x3B; break;
            case 0x37: ctrlcode=0x96; altcode=scancode;
            case 0x39: ctrlcode=0x39; altcode=scancode;
            case 0x47: ctrlcode=0x77; break;
            case 0x48: ctrlcode=0x8D; break;
            case 0x49: ctrlcode=0x84; break;
            case 0x4A: ctrlcode=0x8E; altcode=scancode; break;
            case 0x4B: ctrlcode=0x73; break;
            case 0x4C: ctrlcode=0x8F; break;
            case 0x4D: ctrlcode=0x74; break;
            case 0x4E: ctrlcode=0x90; altcode=scancode; break;
            case 0x4F: ctrlcode=0x75; break;
            case 0x50: ctrlcode=0x91; break;
            case 0x51: ctrlcode=0x76; break;
            case 0x52: ctrlcode=0x92; break;
            case 0x53: ctrlcode=0x93; break;
            case 0x57: scancode=0x85; shiftcode=0x87; ctrlcode=0x89; altcode=0x8B; break;
            case 0x58: scancode=0x86; shiftcode=0x88; ctrlcode=0x8A; altcode=0x8C; break;
            case 0x1D: case 0x2A: case 0x36: case 0x38: case 0x3A:
            case 0x45: case 0x46: case 0x54: case 0x55:
                // modifier keys produce no input
                scancode=0;
        }

        if(event.key.keysym.unicode)
            InputQueue.push_back( std::string(1, char(event.key.keysym.unicode)) );
        else if(scancode)
        {
            if((event.key.keysym.mod & (KMOD_LALT | KMOD_RALT | KMOD_LMETA | KMOD_RMETA)) && altcode)
                scancode = altcode;
            else if((event.key.keysym.mod & (KMOD_LCTRL | KMOD_RCTRL)) && ctrlcode)
                scancode = ctrlcode;
            else if((event.key.keysym.mod & (KMOD_LSHIFT | KMOD_RSHIFT)) && shiftcode)
                scancode = shiftcode;

            std::string code;
            code += '\0';
            code += char(scancode & 0xFF);
            InputQueue.push_back(code);
        }
    }
}

std::string GraphicScreen::Inkeys(bool wait)
{
    CheckInput();

    if(InputQueue.empty()) return "";

    auto val = std::move( InputQueue.front() );
    InputQueue.pop_front();

    return std::move(val);
}

int GraphicScreen::RawInput()
{
    CheckInput();

    if(KeyboardEventQueue.empty()) return LastKeyboardEvent;

    unsigned val = KeyboardEventQueue.front();
    KeyboardEventQueue.pop_front();

    return LastKeyboardEvent = val;
}

