#include <list>
#include <cmath>
#include <map>
#include <deque>
#include <cstring> // memcpy

#include <algorithm>
//#include <SDL_mixer.h>

#include <SDL_timer.h> // for SDL_Delay
#include <SDL_audio.h>

#include <mutex>
#include <memory>
#include <chrono>
#include <thread>

#include "likely.hh"
#include "play.hh"
#include "error.hh"
#include "exec.hh"
#include "events.hh"

#include "Blip_Buffer.h"

void* PlayHandler = 0;
int   PlayThreshold = 0;
enum { TrappingActive=1, TrappingForgotten=-1, TrappingQueued=0 } PlayMode;
int   PlayEventTriggered = 0; // 1-n = lowest-number "remaining notes" trigger that has not been handled yet

static std::map<BAS_LNG, std::size_t> RawHandles;

#include "dbopl.h"


namespace
{
    static const unsigned PreferredFrequency = 44100;
    static const unsigned PreferredChunkSize = 2048;

    static void AudioCallback(void* udata, Uint8* stream, int len);

    void CheckPlayEventTrigger()
    {
        if(PlayMode == TrappingActive && PlayEventTriggered && PlayEventTriggered <= PlayThreshold)
        {
            PlayEventTriggered = 0;
            B_SetEvent(PlayHandler);
        }
    }

    class Mixer
    {
    private:
        bool initialized = false;
        int     output_rate;
        Uint16  output_channels;
        Uint16  output_format;
        unsigned num_active_channels;

        struct Channel
        {
            Mixer& mix;
            std::deque<Uint8> data; // Same format as output.
            std::map<int/*num_notes*/, std::size_t/*where*/> on_play_triggers;

            SDL_AudioCVT convert_input;
            bool mono;
            std::unique_ptr<std::mutex> data_guard;
        public:
            Channel(Mixer& m, int num_channels)
                : mix(m), mono(num_channels==1), data_guard(new std::mutex)
            {
                SDL_BuildAudioCVT(&convert_input,
                    AUDIO_S16SYS,      mono ? 1 : 2,        mix.output_rate,
                    mix.output_format, mix.output_channels, mix.output_rate);
            }
            Channel(const Channel&) = delete;
            Channel(Channel&& b) = default;

            void Feed(const short* ptr, std::size_t num_samples)
            {
                std::size_t num_bytes_input  = num_samples * (mono ? 1 : 2) * 2;
                std::size_t num_bytes_output = num_samples * mix.output_channels * ((mix.output_format & 0xFF) / 8);

                /*fprintf(stderr, "Feeding to channel %p: %u samples; converting %u->%u\n",
                    this, (unsigned)num_samples,
                    (unsigned)num_bytes_input, (unsigned)num_bytes_output);*/

                std::vector<Uint8> output_buffer( std::max(num_bytes_input, num_bytes_output) );
                std::memcpy(&output_buffer[0], ptr, num_bytes_input);
                convert_input.buf = &output_buffer[0];
                convert_input.len = num_bytes_input;
                SDL_ConvertAudio(&convert_input);

                std::lock_guard<std::mutex> channel_lock(*data_guard);
                data.insert(data.end(), output_buffer.begin(), output_buffer.begin() + num_bytes_output);
                SDL_PauseAudio(0);
            }

            void FeedBlank(std::size_t num_bytes_output)
            {
                //fprintf(stderr, "Feeding to channel %p: %u bytes of blank\n", this, (unsigned)num_bytes_output);
                std::lock_guard<std::mutex> channel_lock(*data_guard);
                data.insert(data.end(), num_bytes_output, Uint8(0));
            }

            void OnPlayTriggerHere(int num_notes_remaining)
            {
                std::lock_guard<std::mutex> channel_lock(*data_guard);
                on_play_triggers[num_notes_remaining] = data.size();
            }

            void MixInto(Uint8* stream, std::size_t length, double volume_factor)
            {
                std::lock_guard<std::mutex> channel_lock(*data_guard);
                std::size_t eat = std::min(length, data.size());

                //fprintf(stderr, "Mixing from %p: %u->%u bytes\n", this, (unsigned)data.size(), (unsigned)eat);

                std::vector<Uint8> buffer(eat);
                std::copy(data.begin(), data.begin() + eat, &buffer[0]);

                SDL_MixAudio(stream, &buffer[0], eat, SDL_MIX_MAXVOLUME * volume_factor);

                data.erase(data.begin(), data.begin() + eat);
                for(auto i = on_play_triggers.begin(); i != on_play_triggers.end(); )
                    if(i->second > eat)
                        ++i;
                    else
                    {
                        if(PlayMode == TrappingForgotten)
                            PlayEventTriggered = 0;
                        else if(!PlayEventTriggered || PlayEventTriggered > i->first)
                            PlayEventTriggered = i->first;
                        i = on_play_triggers.erase(i);
                    }
            }
        };
        std::map<std::size_t, Channel> channels;
        std::mutex add_remove_channels_guard;

    public:
        Mixer()
        {
        }
        ~Mixer()
        {
/*
            if(initialized) Mix_CloseAudio();
            Mix_Quit();
*/
        }

        unsigned GetRate()
        {
            Initialize();
            return output_rate;
        }

        std::size_t GetMonoChannel()  { return GetNewChannel(1); }

        std::size_t GetStereoChannel() { return GetNewChannel(2); }

        void AddSamples(std::size_t channel, const short* data, std::size_t num_samples)
        {
            auto i = channels.find(channel);
            if(unlikely(i == channels.end()))
                B_ERROR(51); // Internal error

            i->second.Feed(data, num_samples);
        }

        void OnPlayTriggerHere(std::size_t channel, int num_notes_remaining)
        {
            // Attaches an "ON PLAY()" trigger to the current playing position.
            auto i = channels.find(channel);
            if(unlikely(i == channels.end())) B_ERROR(51); // Internal error
            i->second.OnPlayTriggerHere(num_notes_remaining);
        }

        void Fill(Uint8* stream, int length) // SDL audio callback
        {
            std::lock_guard<std::mutex> channel_lock(add_remove_channels_guard);
            /*fprintf(stderr, "Attempting to fill %d bytes\n", length);*/
            for(auto& c: channels)
                c.second.MixInto(stream, length, 1 / double(channels.size()));
        }

        std::size_t GetMaxQueueBytes() /*const*/
        {
            std::lock_guard<std::mutex> channel_lock(add_remove_channels_guard);
            std::size_t result = 0;
            for(const auto& c: channels)
                result = std::max(result, c.second.data.size());
            return result;
        }

        double GetMaxQueueSeconds()
        {
            unsigned bytes_per_sec = output_channels * output_rate * (output_format & 0xFF) / 8;
            return GetMaxQueueBytes() / double(bytes_per_sec);
        }

    private:
        std::size_t GetNewChannel(int num_channels)
        {
            Initialize();

            std::lock_guard<std::mutex> channel_lock(add_remove_channels_guard);

            std::size_t next_num = 0;
            auto i = channels.rbegin();
            if(i != channels.rend()) next_num = i->first + 1;

            std::pair<std::size_t,Channel> e(next_num, Channel(*this,num_channels));
            auto ip = channels.emplace( std::move(e) );

            std::size_t max_queue_bytes = 0;
            for(const auto& c: channels)
                max_queue_bytes = std::max(max_queue_bytes, c.second.data.size());

            ip.first->second.FeedBlank(max_queue_bytes);

            return next_num;
        }

        void Initialize()
        {
            if(likely(initialized)) return;

            //std::lock_guard<std::mutex> channel_lock(channel_guard);
            if(initialized) return;

            initialized = true;

            //SDL_Init(SDL_INIT_AUDIO);

            SDL_AudioSpec wanted, obtained;
            wanted.freq     = PreferredFrequency;
            wanted.format   = AUDIO_S16SYS;
            wanted.channels = 2;
            wanted.samples  = 4096;
            wanted.callback = AudioCallback;
            wanted.userdata = NULL;

            if (SDL_OpenAudio(&wanted, &obtained) < 0)
            {
                std::fprintf(stderr, "Couldn't open audio: %s\n", SDL_GetError());
                B_ERROR(68); // Device unavailable
            }

            output_rate     = obtained.freq;
            output_channels = obtained.channels;
            output_format   = obtained.format;

            fprintf(stderr, "Audio is %u x %u\n", output_rate,output_channels);
        }

/*
        void InitializeMusicMixer()
        {
            SDL_Init(SDL_INIT_AUDIO);

            Mix_Init(0); // Add MIX_INIT_MP3 etc. if you need them later

            Mix_OpenAudio(PreferredFrequency, AUDIO_S16SYS, 2, PreferredChunkSize);

            int frequency, format, channels;
            if(Mix_QuerySpec(&frequency, &format, &channels))
            {
                if(frequency != PreferredFrequency)
                {
                    // There's no need to try mixing at a different rate than what
                    // the hardware supports.
                    std::fprintf(stderr, "Mixing audio at %d Hz instead of %d Hz.\n",
                        frequency, PreferredFrequency);

                    Mix_CloseAudio();
                    Mix_OpenAudio(frequency, AUDIO_S16SYS, 2, PreferredChunkSize);
                    sampling_rate = frequency;
                }
            }
        }
*/
    };
    Mixer mixer;

    void AudioCallback(void*, Uint8* stream, int len)
    {
        mixer.Fill(stream, len);
    }
}

namespace
{
    bool Background = false;

    static constexpr unsigned long PITrate = 0x1234DD;

    class SquareWavePlayer
    {
    private:
        Blip_Buffer                       Blip;
        Blip_Synth<blip_high_quality,100> Synth;
        blip_time_t time = 0;
        int amplitude = -1;

        unsigned long running_rate;
        std::size_t sound_channel = std::size_t(-1);
    public:
        SquareWavePlayer(unsigned long rate): running_rate(rate)
        {
            // The longest SOUND that can be processed is 65535 ticks,
            // which equals to about 1 hour.
            // For sake of convenience, we quantize it to about 0.05 seconds.
            Blip.set_sample_rate(mixer.GetRate(), 1000);
            Blip.clock_rate(running_rate);
            Synth.output(&Blip);
            Synth.volume(100);
        }

        void SetVolume(int n) // range: 0..100. Default: 100
        {
            Synth.volume(n);
        }

        void FlushAudio()
        {
            long count = Blip.samples_avail();
            if(likely(count <= 0)) return;

            std::vector<blip_sample_t> Audio(count);
            Blip.read_samples(&Audio[0], count);

            mixer.AddSamples(GetMixerHandle(), &Audio[0], Audio.size());;
        }

        void Generate(int freq, int duration) // Duration in PIT ticks: running_rate ticks per second. 0=generate pause.
        {
            int wave_length = duration;
            if(freq)
            {
                wave_length = running_rate / freq;
            }
            /*fprintf(stderr, "Frequency is %d Hz; wave_length is %d. Duration = %d ticks\n",
                freq, wave_length, duration); fflush(stderr);*/

            blip_time_t flush_time = time + running_rate/8;
            blip_time_t end_time   = time + duration;
            int wlhalf = wave_length/2;
            while(time < end_time)
            {
                Synth.update(time, amplitude);

                if(unlikely(time >= flush_time))
                {
                    Blip.end_frame(time);
                    end_time -= time;
                    time = 0;
                    flush_time = time + running_rate/8;
                    FlushAudio();
                }

                time += wlhalf;

                wlhalf = wave_length - wlhalf;
                amplitude = -amplitude;
            }

            Blip.end_frame(end_time);
            time -= end_time;

            FlushAudio();
        }

        std::size_t GetMixerHandle()
        {
            if(unlikely(sound_channel == std::size_t(-1)))
                sound_channel = mixer.GetMonoChannel();

            return sound_channel;
        }
    };

    struct SN76496
    {
        static constexpr unsigned MAX_OUTPUT = 0x7FFF;
        static constexpr unsigned STEP       = 0x10000;
        /* noise feedback for white noise mode (verified on real SN76489 by John Kortink) */
        static constexpr unsigned FB_WNOISE  = 0x14002;   /* (16bits) bit16 = bit0(out) ^ bit2 ^ bit15 */
        /* noise feedback for periodic noise mode */
        static constexpr unsigned FB_PNOISE  = 0x08000;   /* JH 981127 - fixes Do Run Run */
        /* noise generator start preset (for periodic noise) */
        static constexpr unsigned NG_PRESET = 0x0F35;

        int SampleRate;
        unsigned int UpdateStep;
        int VolTable[16];   /* volume table         */
        int Register[8];    /* registers */
        int LastRegister;   /* last register written */
        int Volume[4];      /* volume of voice 0-2 and noise */
        unsigned int RNG;   /* noise generator      */
        int NoiseFB;        /* noise feedback mask */
        int Period[4];
        int Count[4];
        int Output[4];

        void Initialize(unsigned rate)
        {
            /* increase max output basing on gain (0.2 dB per step) */
            double out = MAX_OUTPUT / 3;
            for(int gain=1; gain-- > 0; ) out *= 1.023292992;
            /* build volume table (2dB per step) */
            for (int i = 0;i < 15; ++i)
            {
                /* limit volume to avoid clipping */
                if (out > MAX_OUTPUT / 3) VolTable[i] = MAX_OUTPUT / 3;
                else VolTable[i] = (int)out;
                out /= 1.258925412; /* = 10 ^ (2/20) = 2dB */
            }
            VolTable[15] = 0;
            SampleRate = rate;
            unsigned clock = 3579545;
            UpdateStep = (unsigned int)(((double)STEP * SampleRate * 16) / clock);
            for(unsigned i=0; i<4; ++i)
            {
                Volume[i] = Output[i] = 0;
                Period[i] = Count[i] = UpdateStep;
                Register[i*2+0] = 0;
                Register[i*2+1] = 0x0F;
            }
            RNG       = NG_PRESET;
            NoiseFB   = FB_WNOISE;
            Output[3] = RNG&1;
        }
    };
    class NoisePlayer: private SN76496
    {
        std::size_t sound_channel = std::size_t(-1);
    public:
        void GenerateNoise(int mode, int volume, int duration)
        {
            Initialize();

            Poke(0xE0 + mode);
            Poke(0xFF ^ volume);
            unsigned n_samples = duration * double(SampleRate) / double(PITrate);
            std::vector<short> buffer(n_samples);
            GenerateTo(&buffer[0], n_samples);

            mixer.AddSamples(sound_channel, &buffer[0], n_samples);

            Poke(0xFF);

            // FIXME: We don't emulate the interaction between the third PLAY channel
            //        and NOISE yet.
        }

    private:
        void Initialize()
        {
            if(unlikely(sound_channel == std::size_t(-1)))
            {
                sound_channel = mixer.GetMonoChannel();
                SN76496::Initialize( mixer.GetRate() );
            }
        }

        void Poke(unsigned char data)
        {
            if(data & 0x80)
            {
                int r = (data & 0x70) >> 4, c = r/2;
                LastRegister = r;
                Register[r] = (Register[r] & 0x3F0) | (data & 0x0F);
                switch(r)
                {
                    case 0: case 2: case 4: /* tone frequency */
                        Period[c] = UpdateStep * Register[r];
                        if(Period[c] == 0) Period[c] = 0x3FE;
                        if(r == 4 && (Register[6] & 3) == 3)
                            Period[3] = 2 * Period[2];
                        break;
                    case 1: case 3: case 5: case 7: /* volumes */
                        Volume[c] = VolTable[data & 0x0F];
                        break;
                    case 6: /* noise frequency & mode */
                        int n = Register[6];
                        NoiseFB = (n & 4) ? FB_WNOISE : FB_PNOISE;
                        n &= 3;
                        Period[3] = (n == 3) ? 2 * Period[2] : (UpdateStep << (5+n));
                }
            }
            else
            {
                int r = LastRegister, c = r/2;
                switch(r)
                {
                    case 0: case 2: case 4: /* tone frequency */
                        Register[r] = (Register[r] & 0x0F) | ((data & 0x3F) << 4);
                        Period[c] = UpdateStep * Register[r];
                        if(Period[c] == 0) Period[c] = 0x3FE;
                        if(r == 4 && (Register[6] & 3) == 3)
                            Period[3] = 2 * Period[2];
                }
            }
        }

        void GenerateTo(short* buffer, unsigned length)
        {
            /* Uses:  Count, Period, RNG, NoiseFB, Output
            */
            for(int i=0; i<4; ++i)
            {
                if(Volume[i] == 0)
                    if (Count[i] <= (int)length*(int)STEP) Count[i] += length*STEP;
            }
            for(unsigned count=length; count--; )
            {
                int vol[4], left;
                unsigned int out;

                /* vol[] keeps track of how long each square wave stays */
                /* in the 1 position during the sample period. */
                vol[0] = vol[1] = vol[2] = vol[3] = 0;

                for (unsigned i = 0;i < 3;i++)
                {
                    if (Output[i]) vol[i] += Count[i];
                    Count[i] -= STEP;
                    /* Period[i] is the half period of the square wave. Here, in each */
                    /* loop I add Period[i] twice, so that at the end of the loop the */
                    /* square wave is in the same status (0 or 1) it was at the start. */
                    /* vol[i] is also incremented by Period[i], since the wave has been 1 */
                    /* exactly half of the time, regardless of the initial position. */
                    /* If we exit the loop in the middle, Output[i] has to be inverted */
                    /* and vol[i] incremented only if the exit status of the square */
                    /* wave is 1. */
                    while (Count[i] <= 0)
                    {
                        Count[i] += Period[i];
                        if (Count[i] > 0)
                        {
                            Output[i] ^= 1;
                            if (Output[i]) vol[i] += Period[i];
                            break;
                        }
                        Count[i] += Period[i];
                        vol[i] += Period[i];
                    }
                    if (Output[i]) vol[i] -= Count[i];
                }

                left = STEP;
                do
                {
                    int nextevent;


                    if (Count[3] < left) nextevent = Count[3];
                    else nextevent = left;

                    if (Output[3]) vol[3] += Count[3];
                    Count[3] -= nextevent;
                    if (Count[3] <= 0)
                    {
                        if (RNG & 1) RNG ^= NoiseFB;
                        RNG >>= 1;
                        Output[3] = RNG & 1;
                        Count[3] += Period[3];
                        if (Output[3]) vol[3] += Period[3];
                    }
                    if (Output[3]) vol[3] -= Count[3];

                    left -= nextevent;
                } while (left > 0);

                out = vol[0] * Volume[0] + vol[1] * Volume[1] +
                      vol[2] * Volume[2] + vol[3] * Volume[3];

                if (out > MAX_OUTPUT * STEP) out = MAX_OUTPUT * STEP;

                *(buffer++) = (short)(out / STEP);
            }
        }
    } noise_player;

    class Player: private SquareWavePlayer
    {
    private:
        struct Noterec
        {
            int    note;
            int    volume;
            double duration;
        };
        std::list<Noterec> Buffer;
        bool Triggerer = false;
    public:
        Player() : SquareWavePlayer(PITrate)
        {
        }

        // Duration = length of note in seconds
        void QueueNote(int notenum, int volume, double duration)
        {
            Noterec tmp;
            tmp.note     = notenum;
            tmp.volume   = volume;
            tmp.duration = duration;
            Buffer.push_back(tmp);
            GetMixerHandle();
        }
        void Flush()
        {
            // frequency = (2^(note/12))*440
            // wavelen   = mixer.GetRate() / ((2^(note/12))*440)
            while(!Buffer.empty())
            {
                if(Triggerer)
                    mixer.OnPlayTriggerHere( GetMixerHandle(), Buffer.size() );

                Noterec note = Buffer.front();
                Buffer.pop_front();

                SquareWavePlayer::SetVolume(note.volume);

                //double freq_hz = 440.0 * std::pow(2, (note.note-33) / 12.0);

                // 440 * 2 ^ ((n-33) / 12)
                // 440 * 2 ^ (n/12 - 33/12)
                // 440 * 2 ^ (-33/12) * 2 ^ (n/12)
                // 65.406391325149658665 * 2 ^ (n/12)
                // 65.406391325149658665 * exp(log(2)/12 * log(n))
                if(note.note <= 0 || !note.volume)
                    SquareWavePlayer::Generate(0, note.duration * PITrate);
                else
                {
                    double freq_hz = 65.406391325149658665 * std::exp(0.057762265046662109118*note.note);
                    SquareWavePlayer::Generate(freq_hz, note.duration * PITrate);
                }
            }
        }
        void OnPlayKey()
        {
            // Marks this as a PLAY channel that controls when ON PLAY(n) events are triggered.
            Triggerer = true;
        }
    };

    class PlayParser: public Player
    {
        std::string ParseBuffer;
        size_t      ParsePosition;
        int         Octave;
        int         Tempo;
        int         Length;
        int         Volume;
        int         Rhythm; // n/8
        int         Notenum, Notelength;
    public:
        PlayParser() : ParseBuffer(), ParsePosition(0),
                       Octave(2), Tempo(120), Length(4), Volume(50),
                       Rhythm(7),
                       Notenum(), Notelength()
        {
        }
        void Feed(const std::string& pattern)
        {
            ParseBuffer += pattern;
            Parse();
        }
        void Parse()
        {
/*
  Commands:
    "O" <number>     Sets current octave (0-6)
    "L" <number>     Sets length, L1=whole,  L2=half, ...L64
    "T" <number>     Sets number of quarter notes per minute (32-255, 120 default)
    "V" <number>     Sets volume (0-15)
    "<"              Decrease octave by one
    ">"              Increase octave by one
    <note>
     [#+-]           Explicitly increase or decrease note by one
     [<number>]      Explicit length for that note
     [.]*            Explicitly add half of length (c = 1/1, c.. = 1/1 + 1/2 + 1/4)
    "MB"             Play on background
    "MF"             Play on foreground
    "MS"             Staccato (6/8 note + 2/8 pause)
    "MN"             Normal   (7/8 note + 1/8 pause)
    "ML"             Legato   (8/8 note + 0/8 pause)
    "X" <offset>     Include substring from that address (offset is that returned by VARPTR$(stringvariable))
                      Use VARPTR_ReadStr() from exec.hh
  Numbers:
    [0-9]+           Integer
    "=" <offset>     Include numeric value from that address (offset is that returned by VARPTR$(numvariable))
                      Use VARPTR_ReadNum() from exec.hh
  Notes:
    "N" <number>     Note n (0-84, 0 = rest)
    [ABCDEFG]        Note 0..6 + octave*12
    P                Note 0
*/
            while(ParsePosition < ParseBuffer.size())
            {
                switch(ParseBuffer[ParsePosition])
                {
                    case ' ':
                    case '\t':
                    case '\r':
                    case '\n':
                    case '\v':
                        ++ParsePosition;
                        break;
                    case 'O': case 'o':
                        ++ParsePosition;
                        Octave = (int) ParseNumber();
                        if(Octave < 0 || Octave > 6) default: B_ERROR(5);
                        break;
                    case 'L': case 'l':
                        ++ParsePosition;
                        Length = (int) ParseNumber();
                        if(Length < 1 || Length > 64) B_ERROR(5);
                        break;
                    case 'T': case 't':
                        ++ParsePosition;
                        Tempo = (int) ParseNumber();
                        if(Tempo < 32 || Tempo > 255) B_ERROR(5);
                        break;
                    case 'V': case 'v':
                        ++ParsePosition;
                        Volume = (int) ParseNumber();
                        if(Volume < 0 || Volume > 100) B_ERROR(5);
                        break;
                    case '<':
                        if(Octave > 0) --Octave;
                        ++ParsePosition;
                        break;
                    case '>':
                        if(Octave < 6) ++Octave;
                        ++ParsePosition;
                        break;
                    case 'M': case 'm':
                        ++ParsePosition;
                        if(ParsePosition >= ParseBuffer.size()) B_ERROR(5);
                        switch(ParseBuffer[ParsePosition++])
                        {
                            case 'B': case 'b': Background = true; break;
                            case 'F': case 'f': Background = false; break;
                            case 'S': case 's': Rhythm = 6; break;
                            case 'N': case 'n': Rhythm = 7; break;
                            case 'L': case 'l': Rhythm = 8; break;
                            default: B_ERROR(5);
                        }
                        break;
                    case 'N': case 'n':
                        ++ParsePosition;
                        Notenum = (int) ParseNumber();
                        Notelength = Length;
                        goto GotNote;
                    case 'P': case 'p':
                        ++ParsePosition;
                        Notenum = 0;
                        goto GetOptionalLength;
                    case 'C': case 'c': ++ParsePosition; Notenum = 1 + 0 + Octave*12; goto GetSharpFlat;
                    case 'D': case 'd': ++ParsePosition; Notenum = 1 + 2 + Octave*12; goto GetSharpFlat;
                    case 'E': case 'e': ++ParsePosition; Notenum = 1 + 4 + Octave*12; goto GetSharpFlat;
                    case 'F': case 'f': ++ParsePosition; Notenum = 1 + 5 + Octave*12; goto GetSharpFlat;
                    case 'G': case 'g': ++ParsePosition; Notenum = 1 + 7 + Octave*12; goto GetSharpFlat;
                    case 'A': case 'a': ++ParsePosition; Notenum = 1 + 9 + Octave*12; goto GetSharpFlat;
                    case 'B': case 'b': ++ParsePosition; Notenum = 1 + 11+ Octave*12; goto GetSharpFlat;
                    GetSharpFlat:
                        ParseBlank();
                        while(ParsePosition < ParseBuffer.size())
                            /**/ if(ParseBuffer[ParsePosition] == '+') { ++ParsePosition; ++Notenum; }
                            else if(ParseBuffer[ParsePosition] == '#') { ++ParsePosition; ++Notenum; }
                            else if(ParseBuffer[ParsePosition] == '-') { ++ParsePosition; --Notenum; if(Notenum < 1) B_ERROR(5); }
                            else break;
                    GetOptionalLength:
                        Notelength = (int)ParseNumber(Length);
                    GotNote:
                    {
                        if(Notenum > 84 || Notenum < 0) B_ERROR(5);
                        // tempo  = number of quarter notes in minute
                        // length = 4:quarternote, 8=half of quarternote
                        //
                        //    60 seconds = [tempo] quarternotes
                        //    60 seconds = [tempo] * [length] / 4   notes
                        //    second     = [tempo] * [length] / 240 notes
                        //    note = 240 / ([tempo] * [length]) seconds
                        double duration = (240.0 / 8) / (Tempo * Notelength);
                        double extras = 0, extrashift = 1;
                        ParseBlank();
                        while(ParsePosition < ParseBuffer.size() && ParseBuffer[ParsePosition] == '.')
                        {
                            ++ParsePosition;
                            extras += duration / (extrashift += extrashift);
                            ParseBlank();
                        }
                        if(Rhythm == 8 || Notenum == 0 || Volume == 0)
                            QueueNote(Notenum, Volume, duration * 8);
                        else
                        {
                            QueueNote(Notenum, Volume, duration * Rhythm);
                            QueueNote(0,       Volume, duration * (8-Rhythm) );
                        }
                        break;
                    }
                    case 'X': case 'x':
                    {
                        ++ParsePosition;
                        std::string substring = ParseSubstring();
                        ParseBuffer   = substring + ParseBuffer.substr(ParsePosition);
                        ParsePosition = 0;
                        break;
                    }
                }
            }
        }
    private:
        long double ParseNumber()
        {
            ParseBlank();

            if(ParsePosition >= ParseBuffer.size())
                B_ERROR(5);

            if(ParseBuffer[ParsePosition] == '=')
            {
                if(ParsePosition+(1+sizeof(long)) >= ParseBuffer.size()) B_ERROR(5);
                long double result = VARPTR_ReadNum( &ParseBuffer[ParsePosition] );
                ParsePosition += (1+sizeof(long));
                return result;
            }

            int result = 0, ndigits = 0;
            while(ParsePosition < ParseBuffer.size()
               && ParseBuffer[ParsePosition] >= '0'
               && ParseBuffer[ParsePosition] <= '9')
            {
                result *= 10;
                result += ParseBuffer[ParsePosition++] - '0';
                ++ndigits;
            }

            if(!ndigits) B_ERROR(5);

            return result;
        }

        // When the number is optional
        long double ParseNumber(int defaultvalue)
        {
            ParseBlank();
            if(ParsePosition >= ParseBuffer.size()) return defaultvalue;
            if(ParseBuffer[ParsePosition] == '=')
            {
                if(ParsePosition+(1+sizeof(long)) >= ParseBuffer.size()) B_ERROR(5);
                long double result = VARPTR_ReadNum( &ParseBuffer[ParsePosition] );
                ParsePosition += (1+sizeof(long));
                return result;
            }
            int result = 0, ndigits = 0;
            while(ParsePosition < ParseBuffer.size()
               && ParseBuffer[ParsePosition] >= '0'
               && ParseBuffer[ParsePosition] <= '9')
            {
                result *= 10;
                result += ParseBuffer[ParsePosition++] - '0';
                ++ndigits;
            }
            if(!ndigits) return defaultvalue;
            return result;
        }

        std::string ParseSubstring()
        {
            if(ParsePosition+(1+sizeof(long)) >= ParseBuffer.size()) B_ERROR(5);
            std::string result = VARPTR_ReadStr( &ParseBuffer[ParsePosition] );
            ParsePosition += (1+sizeof(long));
            return result;
        }
        void ParseBlank()
        {
            while(ParsePosition < ParseBuffer.size())
            {
                switch(ParseBuffer[ParsePosition])
                {
                    case ' ':
                    case '\t':
                    case '\r':
                    case '\n':
                    case '\v':
                        ++ParsePosition;
                        break;
                    default:
                        return;
                }
            }
        }
    };

    PlayParser   parsers[3];

    void WaitSound()
    {
        // TODO: Wait until mixer is done playing the samples
        //while(mixer.GetMaxQueueBytes() > mixer.GetRate())
        //    SDL_Delay(50);

        // FIXME: Conflicts with adlib
    }

    void IteratePlay()
    {
        for(size_t a=0; a<3; ++a)
        {
            parsers[a].Flush();
        }

        if(Background)
        {
            // TODO: Add notification when mixer has processed enough samples
            while(mixer.GetMaxQueueBytes() >= mixer.GetRate() * 2)
                SDL_Delay(50);

            // FIXME: Conflicts with adlib
        }
        else
            WaitSound();
    }

    SquareWavePlayer sound_player(PITrate);

    void AdlibMonoAudio(Bitu,Bit32s*);
    void AdlibStereoAudio(Bitu,Bit32s*);
    class AdlibEmulator
    {
    private:
        DBOPL::Handler opl;
        std::size_t sound_channel = std::size_t(-1);
        Bit32u oplregno=0;
        unsigned char oplretval=0;
        std::unique_ptr<std::thread> synthesizer;

    private:
        void Initialize()
        {
            if(unlikely(sound_channel == std::size_t(-1)))
            {
                sound_channel = mixer.GetStereoChannel();
                opl.Init(mixer.GetRate());

                Activate();
            }
        }

        void Activate()
        {
            if(synthesizer) return;
            synthesizer.reset( new std::thread( [this]()
            {
                //std::high_resolution_clock timer;
                std::chrono::high_resolution_clock timer; // For some reason this is in such namespace in GCC.
                //std::system_clock timer;

                long double carry = 0;

                auto previous = timer.now();
                for(;;)
                {
                    SDL_Delay(10); // run 100 times in second

                    auto current = timer.now();
                    auto elapsed = current - previous;
                    auto seconds = elapsed.count() * decltype(elapsed)::period::num
                                     / (long double)(decltype(elapsed)::period::den);
                    long double samples = seconds * mixer.GetRate() + carry;
                    std::size_t num_samples = samples;
                    carry    = samples - num_samples;
                    previous = std::move(current);

                    while(num_samples > 0)
                    {
                        std::size_t n_generated = num_samples;
                        if(n_generated > 512) n_generated = 512; // dbopl limit
                        //fprintf(stderr, "Generates %u adlib samples\n", (unsigned)n_generated);
                        opl.Generate(AdlibMonoAudio, AdlibStereoAudio, n_generated);
                        num_samples -= n_generated;
                    }
                }
            }));
        }
        void Deactivate()
        {
            synthesizer.reset( nullptr );
        }
    public:
        std::size_t GetChannel()  const
        {
            return sound_channel;
        }

        unsigned char Input(unsigned char)
        {
            return oplretval;
        }

        void Output(unsigned char index, unsigned char value)
        {
            Initialize();
            //fprintf(stderr, "OPL(%02X,%02X)\n",index,value);
            switch(index & 3)
            {
                case 0: case 2:
                    oplregno = opl.WriteAddr(index, value);
                    break;
                case 1: case 3:
                    opl.WriteReg( oplregno, value);
                    if(oplregno == 4) { if(value == 0x80) oplretval = 0x02;
                                   else if(value == 0x21) oplretval = 0xC0; }
            }
        }
    } AdlibEmulator;
    void AdlibMonoAudio(Bitu n, Bit32s* src)
    {
        std::vector<short> buffer(n*2);
        for(unsigned a=0; a<n; ++a) buffer[a*2+0] = buffer[a*2+1] = src[a];// / 65536;
        mixer.AddSamples(AdlibEmulator.GetChannel(), &buffer[0], n);
    }
    void AdlibStereoAudio(Bitu n, Bit32s* src)
    {
        std::vector<short> buffer(n*2);
        for(unsigned a=0; a<n*2; ++a) buffer[a] = src[a];// / 65536;
        mixer.AddSamples(AdlibEmulator.GetChannel(), &buffer[0], n);
    }
}

void B_StmtPLAY(const std::string& pattern)
{
    parsers[0].OnPlayKey();
    parsers[0].Feed(pattern);
    IteratePlay();
}

void B_StmtPLAY(const std::string& pattern, const std::string& p2)
{
    parsers[0].OnPlayKey();
    parsers[0].Feed(pattern);
    parsers[1].Feed(p2);
    IteratePlay();
}

void B_StmtPLAY(const std::string& pattern, const std::string& p2, const std::string& p3)
{
    parsers[0].OnPlayKey();
    parsers[0].Feed(pattern);
    parsers[1].Feed(p2);
    parsers[2].Feed(p3);
    IteratePlay();
}

void B_StmtPLAY(int mode)
{
    // 1 = ON, -1 = OFF, 0 = STOP
    switch(mode)
    {
        case -1: PlayMode = TrappingForgotten; break;
        case  1: PlayMode = TrappingActive;    break;
        case  0: PlayMode = TrappingQueued;    break;
    }
    CheckPlayEventTrigger();
}

int B_FuncPLAY()
{
    // FIXME: Return number of notes in queue
    return PlayEventTriggered;
}

void B_StmtSOUND(int freq, int duration/*PIT ticks*/)
{
    if(freq < 1 || freq > long(PITrate/2)) B_ERROR(5); // illegal function call
    // Note: For QBASIC, the lower limit is 37 and upper limit is 32767.
    // We are a bit more relaxed.

    sound_player.Generate(freq, duration * 65536);
    WaitSound();
}

void B_StmtNOISE(int mode/*0..7*/, int volume/*0..15*/, int duration/*PIT ticks*/)
{
    noise_player.GenerateNoise(mode, volume, duration * 65536);
    // Don't WaitSound() here, it's supposed to be queued.
}

void      B_Stmtqb64SNDRAWOPEN(BAS_LNG handle)
{
    auto i = RawHandles.find(handle);
    if(i == RawHandles.end())
        RawHandles.insert( {handle, mixer.GetStereoChannel()} );
}

void      B_Stmtqb64SNDRAW(double left, double right, long handle)
{
    B_Stmtqb64SNDRAWOPEN(handle);

    short samples[2] = { short(32767 * left), short(32767 * right) };
    mixer.AddSamples(RawHandles.find(handle)->second, samples, 1);
}

double    B_Funcqb64SNDRAWLEN() // Return value in seconds
{
    return mixer.GetMaxQueueSeconds();
}

BAS_ULONG B_Funcqb64SNDRATE()
{
    return mixer.GetRate();
}





/* Used by ioemu.cc: */
void B_AdlibOut(unsigned portindex/*0..3*/, unsigned value)
{
    AdlibEmulator.Output(portindex, value);
}

unsigned char B_AdlibIn(unsigned portindex/*0..3*/)
{
    return AdlibEmulator.Input(portindex);
}
